#!/usr/bin/env python

############################################################################
# Copyright (c) 2008,2009,2010,2011 SFB/TR-55
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
# 
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

############################################################################
# BUGS / TODO
# * missing FPGA FIR decoder / minor
# * precisely define number of reply entries for common VPD reads / minor
############################################################################

############################################################################
# qfengine - qpace frontend engine
#
# * fuqd engine
# * multithreading/parallel execution engine
# * flash engine
# * fuqd command assembly
# * qfc decoder engine
############################################################################

import sys
import string
import qlistmod as qq
import copy
import re
import time
import os 
import signal
import subprocess
import select
from threading import Thread
from qfoil import *

############################################################################
# helper functions
############################################################################

# identify the object and give back
# * the identity 'node' / 'psu'
# * the connection rcip / srcip
# * the slot
# * the slot within the next hierarchy
def identify(obj):
    splitstr = string.split(obj, '-')
    if splitstr[0] == 'nc':
        bp    = int(splitstr[1])
        bppos = int(splitstr[2])
        slot  = bppos % 16
        rcip  = 'rc-%02d-%02d' % (bp, bppos / 16)
        return 'node', rcip, bp, slot, bppos
    elif splitstr[0] == 'psu':
        bp    = int(splitstr[1])
        bppos = int(splitstr[2])
        src   = bp / 8
        port  = ((bp % 8) * 3) + bppos
        srcip = 'src-%1d' % (src)
        return 'psu', srcip, bp, port, bppos
    else:
        print 'unknown id %s' % (splitstr[0])
        sys.exit(0)

QFC_KEY_REPLY_DATA = 'reply data'
QFC_KEY_REPLY_LIST = "reply list"

#---------------------------------------------------------------------------
# placeholder class, required for command assembly
class PH(object):
    def __init__(self, name):
        self.name = name
        pass

PH_PORT       = 'ph-port'
PH_SLOT       = 'ph-slot'
PH_NODE_RESET = 'ph-node-reset'
PH_NODE_MUX   = 'ph-node-mux'
PH_NODE_MAC   = 'ph-node-mac'

############################################################################
# engines
############################################################################

# service engine for fuqd
class fuqdEngine(object):

    def __init__(self, list, id='some calls', timeout=240):
        self._connection = False       # connection to fuqd established
        self.targets     = list        # target list
        self.id          = id          # action id
        self.timeout     = timeout     # action timeout
        self.spiverify   = True        # when hitting the combination spi write/read then the data is checked
        self.reinit()                  # all other initialisation

    # re-init
    def reinit(self):
        self._container   = []          # container for commands
        self._action      = None        # fuqd-struct-conversion
        self._gerr        = []          # global errors
        self._daemonreply = {}          # reply of fuqd
        self._emptyreply  = dict.fromkeys(self.targets, { QFC_KEY_LOCAL_ERR : [], QFC_KEY_REPLY_DATA : [] })
        self._unpacked    = dict.fromkeys(self.targets, {}) # unpacked commands, expanded and merged with target list
        self._spiwriteindex = None      # index to last spi write
        self.errdetect    = False       # flag indicating any kind of error
        self.faillist     = []          # list with targets that failed the commands
        self.successlist  = []          # list with targets that passed all commands successfully
        self.reply        = { QFC_KEY_GLOBAL_ERR : self._gerr, \
                              QFC_KEY_REPLY_LIST : self._emptyreply, \
                              QFC_KEY_SUCC_LIST  : self.successlist, \
                              QFC_KEY_FAIL_LIST  : self.faillist }
 
    # commit
    def commit(self):
        self._unpack()
        self._assemble()
        self._connect()
        if not self._gerr:
            self.transceive()
        if not self._gerr:
            self._gerrparse()
        if not self._gerr:
            self._parse()
        self._errcheck()
        self.reply[QFC_KEY_SUCC_LIST] = copy.deepcopy(self.successlist)
        self.reply[QFC_KEY_FAIL_LIST] = copy.deepcopy(self.faillist)

    # check for any error in the reply
    def _errcheck(self):
        if self._gerr: self.errdetect = True
        for target in self.reply[QFC_KEY_REPLY_LIST]:
            if self.reply[QFC_KEY_REPLY_LIST][target][QFC_KEY_LOCAL_ERR]:
                self.faillist.append(target)
        if self.faillist:
            self.errdetect = True
        self.successlist = [obj for obj in self.targets if obj not in self.faillist]

    # push command into container    
    def append(self, obj):
        self._container.append(obj)

    # extend container by several objects
    def extend(self, obj):
        self._container.extend(obj)

    # unpack all commands into the required form
    # this requires to decode all objects in the container
    def _unpack(self):
        # unpack the commands into the dictionary
        for target in self.targets:
            # identify the target
            type, ip, bp, slot, hslot = identify(target)
            
            commands = copy.deepcopy(self._container)
            for cindex in range(len(commands)):
                # look for spi write
                #print commands[cindex][0]
                if commands[cindex][0] in ['FlashExternalWrite', 'FlashInternalWrite']:
                    self._spiwriteindex = cindex
                    #print 'SPI write index', cindex
                # replace all place holders
                for iindex in range(len(commands[cindex])):
                    if isinstance(commands[cindex][iindex], PH):
                        if commands[cindex][iindex].name == PH_SLOT:         # node slot
                            commands[cindex][iindex] = slot
                        elif commands[cindex][iindex].name == PH_PORT:       # psu port
                            commands[cindex][iindex] = slot
                        elif commands[cindex][iindex].name == PH_NODE_MAC:   # nodecard mac address
                            mac = '00 00 ac e0 00 %02x %02x' % (bp, hslot)
                            commands[cindex][iindex] = qq.qdata('', mac)
                        elif commands[cindex][iindex].name == PH_NODE_RESET: # binary formulation for node reset via CPLD
                            mask = ~((1 << slot) & 0xffff)
                            reset = '00 00 %02x %02x' % ((mask >> 8) & 0xff, mask & 0xff)
                            commands[cindex][iindex] = qq.qdata('', reset)
                        elif commands[cindex][iindex].name == PH_NODE_MUX:   # spi muxer
                            mux = '00 00 00 %02x' % (slot)
                            commands[cindex][iindex] = qq.qdata('', mux)

            # create sub dicts for the unpacked calls
            unpacked = {}
            unpacked['TaskList'] = {}
            unpacked['TaskList']['Calls'] = commands
            unpacked['ID'] = target
            unpacked['LockList'] = [ip]

            if type == 'node':
                unpacked['TaskList']['FeenRC'] = ip
            else:
                unpacked['TaskList']['Hemau'] = ip
                unpacked['TaskList']['Slave'] = 1

            self._unpacked[target] = copy.deepcopy(unpacked)

    # embed unpacked commands into the fuqd-like structure
    def _assemble(self):
        self._action = {
            'ID'         : self.id,
            'Timeout'    : self.timeout,
            'GlobalLock' : False,
            'Autoclose'  : False,
            'Persistent' : False,
            'Threads'    : []
        }

        threads = []
        for target in self._unpacked:
            th = {}
            th['LockList'] = self._unpacked[target]['LockList']
            th['ID']       = self._unpacked[target]['ID']
            th['TaskList'] = [self._unpacked[target]['TaskList']]
            threads.append(th)

        self._action['Threads'] = threads

    # connect to daemon
    def _connect(self):
        if self._connection:
            return
        try:
            self.fuqdsock = qq.daemonize("localhost")
        except:
            self._gerr.append('unable to connect to fuqd')
            return
        self._connection = True

    # send and receive xml
    def transceive(self):
        try:
            qq.sendxml(self.fuqdsock, self._action)
        except:
            self._gerr.append('unable to send xml')
            return
        try:
            self._daemonreply = qq.timed_recvxml(self.fuqdsock)
        except:
            self._gerr.append('unable to receive xml')

    # global error parser
    def _gerrparse(self):
        # the root part
        reply = self._daemonreply
        if not reply.has_key('ID'):
            self._gerr.append('missing reply > ID')
        if reply['ID'] != self.id:
            self._gerr.append('wrong reply ID, expected "%s", but received "%s"' % (self.id, reply['ID']))
        if not reply.has_key('errors'):
            self._gerr.append('missing reply > errors')
        if reply['errors']:
            self._gerr.extend(self.reply['errors'])
        if not reply.has_key('Threads'):
            self._gerr.append('missing reply > Threads')
        if self._gerr:
            return

        # the Threads part
        replyth = self._daemonreply['Threads']
        for th in replyth:
            if not th.has_key('ID'):
                self._gerr.append('missing reply > Threads > ID')
                break
            if not th.has_key('errors'):
                self._gerr.append('missing reply > Threads > errors')
                break
            if not th.has_key('ReportList'):
                self._gerr.append('missing reply > Threads > ReportList')
                break

        # done with global error checks

    # now do the local part
    def _parse(self):
        errstr = 'ERROR: in call #%d: %s'

        # error/data parser by reference
        # encoded by 'name' : [data index, error msg index, spi read]
        flib = { 'IICWrite'           : [None, 2, False], \
                 'FlashExternalWrite' : [None, 2, False], \
                 'FlashExternalRead'  : [2,    3, True], \
                 'FlashInternalWrite' : [None, 2, False], \
                 'FlashInternalRead'  : [2,    3, True] }
        # encoded by 'name' : data index
        # src calls don't reply with message on error
        slib = { 'read_direct'        : 2, \
                 'read_linear'        : 2, \
                 'sr_read_word'       : 2, \
                 'srcu_pson'          : None }

        # check all command replies
        replyth = self._daemonreply['Threads']
        for th in replyth:
            target = th['ID']
            #print target
            reply = copy.deepcopy(self.reply[QFC_KEY_REPLY_LIST][target])
            
            # check for errors in the target specific thread
            if th['errors']:
                reply[QFC_KEY_LOCAL_ERR].extend(th['errors'])
            
            report = th['ReportList'][0]
            # check for errors in the target specific ReportList
            if report['errors']:
                reply[QFC_KEY_LOCAL_ERR].extend(report['errors'])

            # even more possible errors *grmpf*
            #if report.has_key('stderr'):
            #    if report['stderr']:
            #        reply[QFC_KEY_LOCAL_ERR].extend(report['errors'])

            # stop parsing if error occured
            if reply[QFC_KEY_LOCAL_ERR]:
                self.reply[QFC_KEY_REPLY_LIST][target] = copy.deepcopy(reply)
                continue

            # now we get the call section and check first for command specific errors
            calls = report['Calls']
            for cindex in range(len(calls)):
                spiread = False
                data = None
                call = calls[cindex]
                c = call[1]

                # service processor commands go separately
                if c  == 'ServiceCommand':
                    if len(call) == 8:
                        reply[QFC_KEY_LOCAL_ERR].append(errstr % (cindex, c))
                        reply[QFC_KEY_LOCAL_ERR].append(call[7])
                        break
                    datatemp = call[5].hexdump(0, newline=False)
                    reply[QFC_KEY_REPLY_DATA].append(string.split(datatemp, ' '))

                # other feenlib calls
                elif c in flib:
                    dindex, eindex, spiread = flib[c]
                    if len(call) == (eindex+1):
                        reply[QFC_KEY_LOCAL_ERR].append(errstr % (cindex, c))
                        reply[QFC_KEY_LOCAL_ERR].append(call[eindex])
                        break
                    if dindex: data = call[dindex]
                    reply[QFC_KEY_REPLY_DATA].append(data)

                # src calls
                elif c in slib:
                    dindex = slib[c]
                    if call[0]:
                        reply[QFC_KEY_LOCAL_ERR].append(errstr % (cindex, c))
                        break
                    if dindex: data = call[dindex]
                    reply[QFC_KEY_REPLY_DATA].append(data)

                # if there is the need to check spi, do it now
                if self.spiverify:
                    if (spiread == True) and (self._spiwriteindex != None):
                        # the original data is still in the container
                        orgdata = self._container[self._spiwriteindex][2]
                        # the read call returns the number of bytes written and also the data
                        if len(orgdata.data) != call[0]:
                            reply[QFC_KEY_LOCAL_ERR].append(errstr % (cindex, c))
                            reply[QFC_KEY_LOCAL_ERR].append('length mismatch on spi data verification')
                        if orgdata.data != call[2].data:
                            reply[QFC_KEY_LOCAL_ERR].append(errstr % (cindex, c))
                            reply[QFC_KEY_LOCAL_ERR].append('data mismatch on spi data verification')
                        #print 'YUM', reply[QFC_KEY_LOCAL_ERR]

            # end analyse calls
            self.reply[QFC_KEY_REPLY_LIST][target] = copy.deepcopy(reply)
        # end analyse threads
    # end parse
# end fuqd engine

#---------------------------------------------------------------------------
# multithreading ssh/scp engine
class pCmdEngine(object):

    def __init__(self, list, id, timeout=10, connecttimeout=3, file=None):
        self.targets     = list        # target list
        self.id          = id          # action id
        self.timeout     = timeout     # action timeout
        self.connecttimeout = connecttimeout     # action timeout
        self.basecmd     = None        # command to execute
        self.file        = file        # bitstream file
        self.parallel    = QFC_PCMD_MAX_PARALLEL
        self.reinit()                  # all other initialisation

    # re-init
    def reinit(self):
        self._action     = None        # fuqd-struct-conversion
        self._gerr       = []          # global errors
        self._emptyreply = dict.fromkeys(self.targets, { QFC_KEY_LOCAL_ERR : [], QFC_KEY_REPLY_DATA : [] })
        self._pcmdreply  = {}          # collect reply of parallel command execution
        self.reply       = { QFC_KEY_GLOBAL_ERR : self._gerr, QFC_KEY_REPLY_LIST : self._emptyreply }
        self.errdetect   = False       # flag indicating any kind of error
        self.faillist    = []          # list with targets that failed the commands
        self.successlist = []          # list with targets that passed all commands successfully
        self.reply       = { QFC_KEY_GLOBAL_ERR : self._gerr, \
                             QFC_KEY_REPLY_LIST : self._emptyreply, \
                             QFC_KEY_SUCC_LIST  : self.successlist, \
                             QFC_KEY_FAIL_LIST  : self.faillist }

    # unpack command
    def _assemble(self):
        if self.id == QFC_CMD_POWER_NODE_PROBE:
            self.basecmd = 'ssh -o ConnectTimeout=%d ? "hostname"' % (self.connecttimeout)
        elif self.id == QFC_CMD_LINKSTRESS:
            self.timeout = 120
            self.connecttimeout = 120
            self.basecmd = 'ssh -o ConnectTimeout=%d ? \'if [ -x "`type -p qfclnkstrss`" ]; then qfclnkstrss || ( echo PASS && logger "qfclnkstrss failed in `hostname`" ); else echo PASS && logger "no have qfclnkstrss on `hostname`"; fi\'' % (self.connecttimeout)
        elif self.id == QFC_CMD_POWER_NODE_REBOOT:
            self.basecmd = 'ssh -o ConnectTimeout=%d ? "reboot"' % (self.connecttimeout)
        elif self.id == QFC_CMD_POWER_NODE_HALT:
            self.basecmd = 'ssh -o ConnectTimeout=%d ? "halt"' % (self.connecttimeout)
        elif self.id == QFC_CMD_FLASH_NODE_LINUX:
            self.basecmd = 'scp -o ConnectTimeout=%d %s ?:/proc/rtas/firmware_update' % (self.connecttimeout, self.file)
            #basecmd += '&& ssh -o ConnectTimeout=%d ? "reboot"' % (self.connecttimeout)
            # check whether file exists are not
            try:
                fd = open(self.file, mode='rb')
                fd.close()
            except:
                self._gerr.append('file error, cannot stat %s' % self.file)
                self.errdetect = True
        else:  
            self._gerr.append('unknown command %s for parallel execution' % self.id)
            self.errdetect = True

    # run all
    def commit(self):
        self._assemble()
        if self._gerr:
            return
        self._pcmd()
        self._parse()
        self._errcheck()
        self.reply[QFC_KEY_SUCC_LIST] = copy.deepcopy(self.successlist)
        self.reply[QFC_KEY_FAIL_LIST] = copy.deepcopy(self.faillist)

    # execute in parallel
    def _pcmd(self):
        # split target list into lists of same size
        p = list2subLists(self.targets, self.parallel)
        #print p
        for l in p:
            self._pcmdreply.update(pSingleCmd(l, self.basecmd, self.timeout))

    # check pcmd reply for errors
    # these lines have taken ~4 hours to get the f*cked up
    # copy operations to work correctly instead of dereferencing
    # the last value of 'stdout'
    def _parse(self):
        if self.id == QFC_CMD_POWER_NODE_PROBE:
            for host in self._pcmdreply:
                reply = copy.deepcopy(self.reply[QFC_KEY_REPLY_LIST][host])
                hostre = re.compile('^%s' % host)
                ret, stdout = self._pcmdreply[host]
                if hostre.match(stdout):
                    reply[QFC_KEY_REPLY_DATA].append('passed')
                else:
                    reply[QFC_KEY_REPLY_DATA].append('failed')
                self.reply[QFC_KEY_REPLY_LIST][host] = copy.deepcopy(reply)

        elif self.id == QFC_CMD_LINKSTRESS:
            for host in self._pcmdreply:
                hostre = re.compile('^PASS')
                ret, stdout = self._pcmdreply[host]
                if hostre.match(stdout):
                    self.reply[QFC_KEY_REPLY_LIST][host][QFC_KEY_REPLY_DATA] = ['passed']
                else:
                    self.reply[QFC_KEY_REPLY_LIST][host][QFC_KEY_REPLY_DATA] = ['failed']
                #end if
            #end for

        elif self.id in [ QFC_CMD_FLASH_NODE_LINUX, QFC_CMD_POWER_NODE_REBOOT, QFC_CMD_POWER_NODE_HALT ]:
            for host in self._pcmdreply:
                reply = copy.deepcopy(self.reply[QFC_KEY_REPLY_LIST][host])
                ret, stdout = self._pcmdreply[host]
                #print host, ret
                if ret == 0:
                    reply[QFC_KEY_REPLY_DATA].append('done')
                else:
                    reply[QFC_KEY_REPLY_DATA].append('failed')
                self.reply[QFC_KEY_REPLY_LIST][host] = copy.deepcopy(reply)

    # check for any error in the reply
    def _errcheck(self):
        if self._gerr: 
            self.errdetect = True
        for target in self.reply[QFC_KEY_REPLY_LIST]:
            if self.reply[QFC_KEY_REPLY_LIST][target][QFC_KEY_REPLY_DATA][0] == 'failed':
                self.faillist.append(target)
        if self.faillist:
            self.errdetect = True
        self.successlist = [obj for obj in self.targets if obj not in self.faillist]

# end pcmdservice

#---------------------------------------------------------------------------
# flash engine for fuqd
# currently instantiates fuqdEngine
# for flash operations
class flashEngine(object):

    def __init__(self, list, id='some calls', file=None, timeout=240):
        self.targets     = list        # target list
        self.id          = id          # action id
        self.timeout     = timeout     # action timeout
        self._file       = file
        self._data       = None        # stream
        self._gerr       = []          # global errors
        self._emptyreply = dict.fromkeys(self.targets, { QFC_KEY_LOCAL_ERR : [], QFC_KEY_REPLY_DATA : [] })
        self.errdetect   = False       # flag indicating any kind of error
        self.faillist    = []          # list with targets that failed the commands
        self.successlist = []          # list with targets that passed all commands successfully
        self.reply       = { QFC_KEY_GLOBAL_ERR : self._gerr, \
                              QFC_KEY_REPLY_LIST : self._emptyreply, \
                              QFC_KEY_SUCC_LIST  : self.successlist, \
                              QFC_KEY_FAIL_LIST  : self.faillist }
 
    # commit
    def commit(self):
        self._unpack()
        if not self._gerr:
            if self.id == QFC_CMD_FLASH_NODE_INTERNAL:
                self._flashinternal()
            else:
                self._flashexternal()
        self._errcheck()
        self.reply[QFC_KEY_SUCC_LIST] = copy.deepcopy(self.successlist)
        self.reply[QFC_KEY_FAIL_LIST] = copy.deepcopy(self.faillist)

    # "unpack"
    def _unpack(self):
        # check whether file exists are not
        try:
            fd = open(self._file, mode='rb')
            self._data = fd.read()
            fd.close()
        except: 
            self._gerr.append('file error, cannot stat %s' % self._file)
            self.errdetect = True

    # flash internal
    def _flashinternal(self):
        # retry counter and node list
        retries = dict.fromkeys(self.targets, 0)
        worklist = copy.deepcopy(self.targets)

        #print 'start flash internal'

        #iteration = 0
 
        # internal flash engine
        while True:

            # if all nodes extracted we're done
            if len(worklist) == 0: break

            #iteration += 1

            #print 'worklist', worklist
            #print 'iteration', iteration

            # the list of nodes we currently work with
            currentlist = extractSingleRootcardNode(worklist)
            
            #print 'currentlist', currentlist

            # first kick the nodes
            reset = fuqdEngine(currentlist, id=QFC_CMD_POWER_NODE_RESET, timeout=self.timeout)
            fuqdCreateCommandCalls(reset, QFC_CMD_POWER_NODE_RESET)
            reset.commit()

            #print 'kicked sp'

            # next try to flash, this does auto-check of the data
            flash = fuqdEngine(currentlist, id=self.id, timeout=self.timeout)
            fuqdCreateCommandCalls(flash, self.id, data=self._data, address=QFC_NODE_FLASH_SP_IMAGE_ADDR)
            flash.commit()
           
            #print 'flash done'
            #print 'flash gerr', flash._gerr
            #print 'flash errdetect', flash.errdetect
            #print 'flash faillist', flash.faillist
            #print 'flash successlist', flash.successlist

            # register retries 
            for node in flash.faillist:
                retries[node] += 1
                #print 'node', node, 'retries', retries[node]
                # check if threshold is reached
                if retries[node] >= QFC_FLASH_RETRY_THRESHOLD_INTERNAL:
                    reply = copy.deepcopy(self.reply[QFC_KEY_REPLY_LIST][node])
                    reply[QFC_KEY_LOCAL_ERR].append('internal flash retry threshold reached')
                    self.reply[QFC_KEY_REPLY_LIST][node] = copy.deepcopy(reply)

            # update the worklist, nodes in success list have passed flash process
            worklist = [item for item in worklist if item not in flash.successlist]
    # end flash internal

    def _flashexternal(self):
        # get the offset into spansi
        #if QFC_METHOD_MAP[self.id] == QFC_CMD_FLASH_NODE_FPGA:
        if self.id == QFC_CMD_FLASH_NODE_FPGA:
            offset = QFC_NODE_FLASH_FPGA_IMAGE_ADDR
        else:
            offset = QFC_NODE_FLASH_SLOF_IMAGE_ADDR 

        # stream must be split into chunks
        chunks = chunkList(len(self._data), QFC_NODE_FLASH_BLOCKSIZE)
        print len(chunks)

        # retry counter and node list
        retries = dict.fromkeys(self.targets, 0)
        worklist = copy.deepcopy(self.targets)

        # a comment on the lists:
        # * worklist contains all nodes that have to be flashed
        # * currentflashlist contains all nodes that are currently in the flash process
        # * sectorflashlist is required to support for sector reflashing 
        while True:

            # done
            if len(worklist) == 0: break

            # each set of nodes start from the beginning
            dataoffset  = 0
            flashoffset = offset

            # get a list with only one node per rootcard
            currentlist = extractSingleRootcardNode(worklist)

            #print 'worklist', worklist
            #print 'currentflashlist', currentflashlist
            # kick the nodes

            # first kick the nodes
            reset = fuqdEngine(currentlist, id=QFC_CMD_POWER_NODE_RESET, timeout=self.timeout)
            fuqdCreateCommandCalls(reset, QFC_CMD_POWER_NODE_RESET)
            reset.commit()

            # try to flash all chunks for the nodes in the current flash list
            for chunk in range(len(chunks)):

                # data preparation for the current chunk
                chunksize = chunks[chunk]
                dataslice = self._data[dataoffset:(dataoffset+chunksize)]
                sectorlist = copy.deepcopy(currentlist)
                print 'sectorlist', sectorlist
                print 'chunk', chunk

                # repeat flashing one sector until ok or threshold reached
                while True: 

                    # this sector done
                    if len(sectorlist) == 0: break

                    # next try to flash, this does auto-check of the data
                    flash = fuqdEngine(sectorlist, id=self.id, timeout=self.timeout)
                    fuqdCreateCommandCalls(flash, self.id, address=flashoffset, data=dataslice)
                    flash.commit()

                    for node in flash.faillist:
                        print node, 'sector flash failed'
                        retries[node] += 1
                        # check if threshold is reached, remove those nodes from sector flash list
                        if retries[node] >= QFC_FLASH_RETRY_THRESHOLD_EXTERNAL:
                            reply = copy.deepcopy(self.reply[QFC_KEY_REPLY_LIST][node])
                            reply[QFC_KEY_LOCAL_ERR].append('external flash retry threshold reached')
                            self.reply[QFC_KEY_REPLY_LIST][node] = copy.deepcopy(reply)
                            sectorlist.remove(node)
                            currentlist.remove(node)
                            worklist.remove(node)
                    # update sector list
                    sectorlist = [item for item in sectorlist if item not in flash.successlist]
                # done single sector
                dataoffset  += chunksize
                flashoffset += chunksize
            # end flashing nodes in current flash list, update the worklist
            worklist = [item for item in worklist if item not in currentlist]
    # end flash internal

    # check for any error in the reply
    def _errcheck(self):
        if self._gerr: 
            self.errdetect = True
        for target in self.reply[QFC_KEY_REPLY_LIST]:
            if self.reply[QFC_KEY_REPLY_LIST][target][QFC_KEY_LOCAL_ERR]:
                self.faillist.append(target)
        if self.faillist:
            self.errdetect = True
        self.successlist = [obj for obj in self.targets if obj not in self.faillist]

# end flashEngine

#---------------------------------------------------------------------------
class thd(Thread):
    def __init__ (self, cmd, timeout):
        Thread.__init__(self)
        self.cmd = cmd 
        self.timeout = timeout
        self.blocksize = 4096         # standard page size for read buffer
        self.stdout = ''              # stdout buffer
        self.returncode = 0           # exit code
        self.killed = False           # true when if, false if
        self.killtry = 0              # number of os kills
    def run(self):
        starttime = time.time()
        p = subprocess.Popen(self.cmd, 
                             shell=True,
                             bufsize=2*self.blocksize, 
                             stdin=subprocess.PIPE, 
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE )
        po = p.stdout.fileno()
        #pe = p.stderr.fileno()
        while p.poll() is None:
            if time.time() - starttime > self.timeout:  # timed out
                self.killtry += 1 
                if self.killtry == 1:  # try a SIGTERM first
                    self.killed = True
                    try:
                        os.kill(p.pid, signal.SIGTERM)
                    except: #OSerror:
                        pass
                else:  # try a SIGKILL only when not done yet!
                    try:
                        os.kill(p.pid, signal.SIGKILL)
                    except: #OSerror:
                        pass
                time.sleep(1)
            else:  # wait for data
                #( so, si, se ) = select.select( [po], [], [pe], 1 )  # wait max 1 sec for something to happen on p.stdout
                (so, si, se) = select.select([po], [], [], 1)  # wait max 1 sec for something to happen on p.stdout
                if so:
                    self.stdout += str(p.stdout.read(self.blocksize))
                #if se:
                #    print p.stderr.read( self.blocksize )
        (so, si, se) = select.select([po], [], [], 1)  # this final read is required, otherwise data is lost for ping!
        if so:
             self.stdout += str( p.stdout.read(self.blocksize))
        self.returncode = p.wait()

#---------------------------------------------------------------------------
# run a single command in parallel on all hosts
# hosts is a list
# ? in cmd will be replaced by the hostname
# returns dictionary [ "host" : [exitcode, stdout] ]
def pSingleCmd(hosts, cmd, timeout=5):
    threadlist = []
    r = {}
    # spawn threads
    for host in hosts:
        tcmd = cmd.replace('?', host)
        pthread = thd(tcmd, timeout)
        threadlist.append([host, pthread])
        pthread.start()

    # join em
    for host, pthread in threadlist:
        pthread.join()
        try:
            rv = int(pthread.returncode)  # convert to int
        except:
            rv = -1                       # didn't work

        # make the return dict
        r[host] = [rv,pthread.stdout];

    return r

############################################################################
# service kernel functions
############################################################################

# fuqdVpdRead is used to create read call(s)
# for data in VPD
def fuqdVpdRead(srv, address, length):

    chunks = chunkList(length, QFC_NODE_VPD_BLOCKSIZE)
    offset = address
    for chunk in chunks:
        spcmd = '%02x %02x %02x' % ((offset & 0xff00) >> 8, offset & 0xff, chunk)
        payload = qq.qdata('', spcmd)
        srv.append(['ServiceCommand', PH(PH_SLOT), QFC_NODE_SP_VPD_READ, payload])
        offset += chunk

#---------------------------------------------------------------------------
# fuqdVpdWrite is used to create write call(s)
# for data in VPD. data is expected to be a list
# of qdata-convertible 2-byte list entries of
# the form ['xx', 'xx', ...]
def fuqdVpdWrite(srv, address, data):

    datalist = list2subLists(data, QFC_NODE_VPD_BLOCKSIZE)
    offset = 0
    for sublist in datalist:
        payloadstr = binaryStr2qdataStr(sublist)
        addr = address + offset
        spcmd = '%02x %02x %02x %s' % ((addr & 0xff00) >> 8, addr & 0xff, len(sublist), payloadstr)
        payload = qq.qdata('', spcmd)
        srv.append(['ServiceCommand', PH(PH_SLOT), QFC_NODE_SP_VPD_WRITE, payload])
        offset += len(sublist)

#---------------------------------------------------------------------------
# fuqdCreateCommandCalls creates a list of calls
# that have to executed depending on the method
def fuqdCreateCommandCalls(srv, ocommand, address=None, length=None, data=None):

    # abbreviations for placeholders
    slot   = PH(PH_SLOT)
    port   = PH(PH_PORT)
    reset  = PH(PH_NODE_RESET)
    spimux = PH(PH_NODE_MUX)
    mac    = PH(PH_NODE_MAC)

    # translate the ocommand into a reduced command for code re-usage
    # expand tail if existing
    command, tail = QFC_METHOD_MAP[ocommand]
    if tail:
        address, length, data = tail

    # --- define call sequences ---

    print 'original command', ocommand
    print 'creating command', command
    print 'address', address
    print 'length ', length
    #print 'data   ', data

    # get
    if command == QFC_CMD_GET_NODE_TEMP:
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_TEMPERATURE, qq.qdata('', 'ff')])

    elif command == QFC_CMD_CLEAR_NODE_STATES:
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CLEAR_STATES, qq.qdata('', '')])
        fuqdVpdWrite(srv, QFC_NODE_VPD_FIR_HEADER_ADDR,  ['00', '00', '00', '00'])
        fuqdVpdWrite(srv, QFC_NODE_VPD_FPGA_FIR_ADDR,    ['00', '00', '00', '00'])

    elif command == QFC_CMD_GET_NODE_VERSION:
        srv.append(['IICWrite', 0x05, spimux]) # data
        srv.append(['IICWrite', 0x06, qq.qdata('', '01')])  # addr
        srv.append(['IICWrite', 0x07, qq.qdata('', '01')])  # command
        srv.append(['FlashExternalRead', QFC_NODE_FLASH_FPGA_VERSION_ADDR, QFC_NODE_FLASH_FPGA_VERSION_LEN])
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_VERSION, qq.qdata('', '')])
        srv.append(['FlashExternalRead', QFC_NODE_FLASH_SLOF_VERSION_ADDR, QFC_NODE_FLASH_SLOF_VERSION_LEN])

    elif command == QFC_CMD_GET_NODE_VPD:
        fuqdVpdRead(srv, address, length)

    elif command == QFC_CMD_GET_NODE_STATUS:
        # first the CPLD stuff
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CPLD_READ, qq.qdata('', '00 ff')])   # Port 0x00 -> ncStatus
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CPLD_READ, qq.qdata('', '01 ff')])   # Port 0x01 -> ncPowerSeqStatus
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CPLD_READ, qq.qdata('', '02 ff')])   # Port 0x02 -> ncSomaVddStatus
        # it's faster to create a list for all the status events
        statii = ['00', '02', '03', '04', '05', '06', '07', '08', '09', '0c', '10', '11', '12', '13', '14', '15']
        sp = QFC_NODE_SP_STATUS_CHANGE
        for stat in statii:
            srv.append(['ServiceCommand', slot, sp, qq.qdata('', stat)])
                     
    elif command == QFC_CMD_GET_NODE_CONFIG:
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CPLD_READ, qq.qdata('', '09 ff')])      # Port 0x09 -> ncPowerSeqConfig
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_VPD_READ,  qq.qdata('', '00 14 02')])   # VPD -> BootFlags, RracClk
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CPLD_READ, qq.qdata('', '0a ff')])      # Port 0x0a -> PMC ref clk
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CPLD_READ, qq.qdata('', '0c ff')])      # Port 0x0c -> clock source

    # careful: this only works with a blocksize of 250 bytes per read, otherwise decoding has to be adopted!
    elif command == QFC_CMD_GET_NODE_DUMP:
        fuqdVpdRead(srv, QFC_NODE_VPD_BE_FIR_ADDR,   QFC_NODE_VPD_BE_FIR_LEN)
        fuqdVpdRead(srv, QFC_NODE_VPD_FPGA_FIR_ADDR, QFC_NODE_VPD_FPGA_FIR_LEN)

    elif command == QFC_CMD_GET_PSU_STATUS:
        srv.append(['srcu_init', port])
        srv.append(['sr_read_word', port, QFC_PSU_PMBUS_STATUS_WORD])
        srv.append(['read_direct',  port, QFC_PSU_PMBUS_VOUT, 'T'])
        srv.append(['read_direct',  port, QFC_PSU_PMBUS_VOUT_COMMAND, 'T'])
        srv.append(['read_linear',  port, QFC_PSU_PMBUS_VOUT_VIN, 'T'])
        srv.append(['read_direct',  port, QFC_PSU_PMBUS_VOUT_IOUT, 'T'])
        srv.append(['read_linear',  port, QFC_PSU_PMBUS_VOUT_IIN, 'T'])
        srv.append(['read_linear',  port, QFC_PSU_PMBUS_VOUT_POUT, 'T'])
        srv.append(['read_linear',  port, QFC_PSU_PMBUS_VOUT_PIN, 'T'])
        srv.append(['read_linear',  port, QFC_PSU_PMBUS_TEMPERATURE_1, 'T'])
        srv.append(['read_direct',  port, QFC_PSU_PMBUS_TEMPERATURE_2, 'T'])
        srv.append(['read_linear',  port, QFC_PSU_PMBUS_FAN_SPEED_1, 'F'])
        srv.append(['read_linear',  port, QFC_PSU_PMBUS_FAN_SPEED_2, 'F'])

    # set
    elif command == QFC_CMD_SET_NODE_MAC:
        if data:
            macstr  = '00 %s %s %s %s %s %s' % (data[0], data[1], data[2], data[3], data[4], data[5])
            cmd     = ['ServiceCommand', slot, QFC_NODE_SP_SET_NODE_ID, qq.qdata('', macstr)]
        else:
            cmd     = ['ServiceCommand', slot, QFC_NODE_SP_SET_NODE_ID, mac]
        srv.append(cmd)

    elif command == QFC_CMD_SET_NODE_VPD:
        fuqdVpdWrite(srv, address, data)

    elif command == QFC_CMD_SET_NODE_CLOCK:
        clkstr = '0c 04 %s' % data[0]
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_CPLD_WRITE, qq.qdata('', clkstr)])   # Port 0x0c, bit 4 -> clock source

    # power
    elif command == QFC_CMD_POWER_NODE:
        srv.append(['ServiceCommand', slot, QFC_NODE_SP_POWER, qq.qdata('', data[0])])

    elif command == QFC_CMD_POWER_NODE_RESET:
        srv.append(['IICWrite', 0x05, reset]) # data
        srv.append(['IICWrite', 0x06, qq.qdata('', '20')])  # addr
        srv.append(['IICWrite', 0x07, qq.qdata('', '01')])  # command
        srv.append(['IICWrite', 0x05, qq.qdata('', '00 00 ff ff')]) # data
        srv.append(['IICWrite', 0x06, qq.qdata('', '20')])  # addr
        srv.append(['IICWrite', 0x07, qq.qdata('', '01')])  # command

    elif command == QFC_CMD_POWER_PSU:
        srv.append(['srcu_init', port])
        srv.append(['srcu_pson', port, data[0]])

    # flash
    elif command == QFC_CMD_FLASH_NODE_INTERNAL:
        srv.append(['IICWrite', 0x05, spimux]) # data
        srv.append(['IICWrite', 0x06, qq.qdata('', '01')])  # addr
        srv.append(['IICWrite', 0x07, qq.qdata('', '01')])  # command
        srv.append(['FlashInternalWrite', 0, qq.qdata(data)])
        srv.append(['FlashInternalRead', 0, len(data)])
        #srv.append(['ServiceCommand', slot, QFC_NODE_SP_POWER, qq.qdata('', '03')])

    elif command == QFC_CMD_FLASH_NODE_EXTERNAL:
        srv.append(['IICWrite', 0x05, spimux]) # data
        srv.append(['IICWrite', 0x06, qq.qdata('', '01')])  # addr
        srv.append(['IICWrite', 0x07, qq.qdata('', '01')])  # command
        srv.append(['FlashExternalWrite', address, qq.qdata(data)])
        srv.append(['FlashExternalRead', address, len(data)])

    # end if
# end fuqdCreateCommandCalls

#---------------------------------------------------------------------------
# unify access to all engines
def unifiedRequest(list, command, address=None, length=None, data=None, file=None):

    engine = None

    # check the command for the appropriate engine
    if command in QFC_ENGINE_FUQD_LIST:
        engine = fuqdEngine(list, id=command)

    elif command in QFC_ENGINE_PCMD_LIST:
        engine = pCmdEngine(list, id=command, file=file)

    elif command in QFC_ENGINE_FLASH_LIST:
        engine = flashEngine(list, id=command, file=file)

    # engine consistency check
    if not engine:
        print 'unable to map engine. exiting.'
        sys.exit(1)

    # define the call sequence for fuqd requests
    if command in QFC_ENGINE_FUQD_LIST:
        fuqdCreateCommandCalls(engine, command, address=address, length=length, data=data)

    # run
    engine.commit()

    return engine.reply

###########################################################################
# service evaluation functions
############################################################################

# display/evaluate reply of common nodecard information
# note: place commands that require special
#       kind of evaluation before evaluation
#       of local errors
def evalCommonCmdReply(reply, command, address=0, silent=False, grep=False, short=False):

    if short:
        grep = True

    if silent:
        devnull = open("/dev/null","w")
        oldout  = sys.stdout
        sys.stdout = devnull

    faillist = []

    objreply = reply[QFC_KEY_REPLY_LIST]
    objkeys = objreply.keys()
    objkeys.sort()

    gerr = reply[QFC_KEY_GLOBAL_ERR]

    print '\n'

    # global error handling
    if gerr:
        print '=' * QFC_OUTPUT_BOARDER_SIZE
        print 'GLOBAL ERRORS:'
        for gerr in reply[QFC_KEY_GLOBAL_ERR]:
            print '  *', gerr
        print '=' * QFC_OUTPUT_BOARDER_SIZE
        faillist = objkeys

    if faillist:
        if silent:
            sys.stdout = oldout
            devnull.close()
        return faillist

    key = QFC_KEY_REPLY_DATA

    if short:
        print '=' * QFC_OUTPUT_BOARDER_SIZE
        print 'command: %s' % (command)
        print '-' * QFC_OUTPUT_BOARDER_SIZE

    for obj in objkeys:
        objdata = objreply[obj]
        lerr = objdata[QFC_KEY_LOCAL_ERR]
        if grep: g = '%s::' % obj 
        else:    g = ''
        if not short:
            print '=' * QFC_OUTPUT_BOARDER_SIZE
            print '%s (%s)' % (obj, command)
            print '-' * QFC_OUTPUT_BOARDER_SIZE

        if command == QFC_CMD_POWER_NODE_PROBE:
            data = objdata[key]
            print g+'ncSshProbe          =', data[0],

        elif command in [ QFC_CMD_FLASH_NODE_LINUX, QFC_CMD_POWER_NODE_REBOOT, QFC_CMD_POWER_NODE_HALT ]:
            data = objdata[key]
            print g+'%s' % (data[0]),

        elif len(lerr) > 0:
            faillist.append(obj)
            print g+'LOCAL ERRORS:'
            for err in lerr:
                print g+'  *', err

        elif command == QFC_CMD_GET_NODE_TEMP:
            print g+'ncCurrentTemp       =',
            for temp in objdata[key][0]:
                print int(temp, 16),

        # note: first 3 commands are iiw replies
        # note: spi read replies with binary data
        # decoder functions work with split hexdump representation
        elif command == QFC_CMD_GET_NODE_VERSION:
            nwpbin  = objdata[key][3].hexdump(0, newline=False)
            nwphex  = string.split(nwpbin, ' ')
            slofbin = objdata[key][5].hexdump(0, newline=False)
            slofhex = string.split(slofbin, ' ')
            print g+'ncNwpRev            =', fpgaIdentifyBitstream(nwphex)
            print g+'ncSpFirmwareRev     =', qdataStringify(objdata[key][4])
            print g+'ncSlofRev           =', qdataStringify(slofhex),

        elif command == QFC_CMD_GET_NODE_SERIAL:
            print g+'ncSerial            =', qdataStringify(objdata[key][0]),

        elif command == QFC_CMD_GET_NODE_MAC:
            mac = objdata[key][0]
            macstr = '%s:%s:%s:%s:%s:%s' % (mac[0], mac[1], mac[2], mac[3], mac[4], mac[5])
            print g+'ncMacAddr           =', macstr,

        elif command == QFC_CMD_GET_NODE_VPD:
            vpddata = objdata[key]
            # the vpd data comes in chunks, so we must connect them
            data = []
            for d in vpddata:
                data.extend(d)
            # ready to decode
            padding = address % 16
            padded = binaryStrPad(data, padding)
            full = len(padded) / 16
            vpdoff = address - padding
            dataoff = 0
            for line in range(full):
                print g+'0x%04x --' % (vpdoff),
                for i in range(16):
                    print padded[dataoff + i],
                vpdoff += 16
                dataoff += 16
                if (line != full - 1): print ''

        elif command == QFC_CMD_GET_NODE_STATUS:
            data = objdata[QFC_KEY_REPLY_DATA]
            d = int(data[0][0], 16)
            print g+'ncStatus            = 0x%02x %s' % (d, ncStatusDecode(d))
            d = int(data[1][0], 16) & 0x0f
            print g+'ncPowerSeqStatus    = 0x%02x %s' % (d, ncPowerSeqStatusDecode(d))
            d = int(data[2][0], 16) & 0x3f
            print g+'ncSomaVddStatus     = 0x%02x %s' % (d, ncSomaVddStatusDecode(d))
            d = int(data[3][0], 16)
            print g+'Last status change  = #%02d %s' % (d, ncStatusEventDecode(d))
            d = binaryStr2quad(data[4])
            print g+'-- #02 Boot cycle counter       = %d' % (d[0])
            d = int(data[5][0], 16)
            print g+'-- #03 Boot status              = 0x%02x %s' % (d, ncBootStatusDecode(d))
            d = int(data[6][0], 16)
            print g+'-- #04 CPU thread fault, ID     = 0x%02x' % (d)
            d = binaryStr2int(data[7])
            print g+'-- #05 DIMM fault               = 0x%02x%02x on DIMM %02d' % (d[0], d[1], d[2])
            d = int(data[8][0], 16)
            print g+'-- #06 VDD fault mask           = 0x%02x' % (d)
            d = binaryStr2int(data[9])
            print g+'-- #07 Temp. sensor status      = 0x%02x (%03u %03u %03u %03u %03u %03u %03u)' % \
                (d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7])
            d = int(data[10][0], 16)
            print g+'-- #08 Soma thermal alert       = 0x%02x (0 no alert, 1 alert detected)' % (d)
            d = int(data[11][0], 16)
            print g+'-- #09 Checkstop                = 0x%02x (1 Cell, 5 FPGA)' % (d)
            d = int(data[12][0], 16)
            print g+'-- #12 Watchdog expired         = 0x%02x (0xff not expired)' % (d)
            d = binaryStr2int(data[13])
            print g+'-- #16 Host checkpoint          = 0x%02x%02x (0xffff no checkpoint set)' % (d[0], d[1])
            d = int(data[14][0], 16)
            print g+'-- #17 POR error code           = 0x%02x %s' % (d, ncPorDecode(d))
            d = binaryStr2quad(data[15])
            print g+'-- #18 Boot failure counter     = %d' % (d[0])
            d = binaryStr2quad(data[16])
            print g+'-- #19 Recoverable FIR          = 0x%08x' % (d[0])
            d = binaryStr2quad(data[17])
            print g+'-- #20 CPU attention            = 0x%08x' % (d[0])
            d = binaryStr2quad(data[18])
            print g+'-- #21 FPGA check               = 0x%08x' % (d[0]),

        elif command == QFC_CMD_GET_NODE_DUMP:
            vpddata = objdata[key]  # BE dump
            # the vpd data comes in chunks, so we must connect them
            data = []
            for d in vpddata:
                data.extend(d)
            # ready to decode
            quad = binaryStr2quad(data)
            print g+'BE FIR hexdump (valid from offset 0x0 - 0x2a4, VPD base 0x2010):'
            print g
            print g+'   Header/cntr     0x%08x  0x%08x' % (quad[0], quad[1])
            print g+'   Timestamp       %s' % (time.asctime(time.gmtime(quad[2])))
            print g
            print g+'   BE FIR dump: address - hex -- bin'
            print g
            firoff = 0
            dataoff = 4
            for line in range(85):
                print g+'   0x%04x - 0x%08x %08x -- %s %s' % \
                    (firoff, quad[dataoff], quad[dataoff+1], quad2binary(quad[dataoff]), quad2binary(quad[dataoff+1]))
                firoff += 8
                dataoff += 2
            print g
            print g+"BE FIR check 'checkstop_fir':   0x%08x -- %s" % (quad[15], ncBEglobalFirDecode(quad[15]))
            print g+"BE FIR check 'recoverable_fir': 0x%08x -- %s" % (quad[16], ncBErecoverableFirDecode(quad[16]))
            print g
            if quad[0] == 0x00c3ec3e: 
                print g+'BE FIR dump check: dump valid!',
            else: 
                print g+'BE FIR dump check: dump invalid! (header did not match 0x00c3ec3e)',
            
            # TODO: implement FPGA Fir decoder

        elif command == QFC_CMD_GET_NODE_CONFIG:
            data = objdata[key]
            d = int(data[0][0], 16) & 0x1f
            print g+'ncPowerSeqConfig    = 0x%02x %s' % (d, ncPowerSeqConfigDecode(d))
            d = int(data[1][0], 16)
            print g+'ncBootFlags         = 0x%02x %s' % (d, ncBootFlagDecode(d))
            d = int(data[1][1], 16)
            print g+'ncRracClk           = 0x%02x %s' % (d, ncRracClkDecode(d))
            d = (int(data[2][0], 16) >> 6) & 0x1
            s = ['(ref_clk 156 MHz)', '(ref_clk 125 MHz)']
            print g+'ncClkSelect         = 0x%02x %s' % (d, s[d])
            d = (int(data[3][0], 16) >> 4) & 0x1
            s = ['(on-board)', '(backplane)']
            print g+'ncTnwClockSource    = 0x%02x %s' % (d, s[d]),

        elif command == QFC_CMD_GET_PSU_STATUS:
            data = objdata[key]
            print g+'psuStatus               = 0x%04x %s' % (data[0], psuStatusDecode(data[0]))
            print g+'psuCommandOutputVoltage =%9.2f V' % (data[1])   
            print g+'psuOutputVoltage        =%9.2f V' % (data[2])   
            print g+'psuInputVoltage         =%9.2f V' % (data[3])   
            print g+'psuOutputCurrent        =%9.2f A' % (data[4])   
            print g+'psuInputCurrent         =%9.2f A' % (data[5])   
            print g+'psuOutputPower          =%9.2f W' % (data[6])   
            print g+'psuInputPower           =%9.2f W' % (data[7])   
            print g+'psuTemperature1         =%9.2f C' % (data[8])   
            print g+'psuTemperature2         =%9.2f C' % (data[9])   
            print g+'psuFanSpeed1            =%9.2f RPM' % (data[10]) 
            print g+'psuFanSpeed2            =%9.2f RPM' % (data[11]), 

        else: 
            print g+'done',

        if not short:
            print '\n'
        else:
            print ''

    print '\n'

    if silent:
        sys.stdout = oldout
        devnull.close()

# end eval
